#!/usr/bin/env python
import sys
import whrandom
import pickle

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Library General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

XSIZE = 100
YSIZE = 100

MOUNTAIN_PERCENT = 5
HILL_PERCENT = 5
WATER_PERCENT = 10
FOREST_PERCENT = 10
SWAMP_PERCENT = 3
CITY_PERCENT = 0.8
RUIN_PERCENT = 0.25
TEMPLE_PERCENT = 0.03
CITY_FAR_APART = 3
RUIN_FAR_APART = 4
TEMPLE_FAR_APART = 10

MOUNTAIN = 'M'
HILL = 'h'
WATER = '~'
FOREST = '$'
PLAIN = '.'
SWAMP = '_'
CITY = 'C'
RUIN = 'r'
TEMPLE = 'T'
NOTHING = ' '
PART_OF_CITY = 'c'

"""
Maps in this code refer to a variety of things. In Python, they are just
two dimensional arrays, but in other languages, they would be 2D arrays of
booleans, ints, or chars. Many of the map functions are generic and don't
care what they're operating on.

Author: John Farrell friendless@users.sourceforge.net
"""

def toStringMap(map):
    "Convert a map to a list of strings."
    xsize = len(map)
    ysize = len(map[0])
    result = []
    stringType = type('string')
    for y in range(ysize):
        s = ''
        for x in range(xsize):
            v = map[x][y]
            if type(v) == stringType:
                s = s + v
            else:
                s = s + `v`
        result.append(s)
    return result
    
"""Print a result of toStringMap"""
def writeToFile(file, map):
    for s in map:
        file.write(s)
        
def genmap(func):
    "Generate a new map."
    result = []
    for col in range(XSIZE):
        c = []
        for row in range(YSIZE):
            c.append(func(col, row))
        result.append(c)
    return result    
    
def sum(map):
    "Return the total of a map of integers."
    count = 0
    xsize = len(map)
    ysize = len(map[0])
    for y in range(ysize):
        for x in range(xsize):
            count = count + map[x][y] 
    return count 
            
def size(map):
    "Total number of positions in map"
    xsize = len(map)
    ysize = len(map[0])
    return xsize * ysize  
    
def countValues(map, allocated):
    """
        Count the number of each value in a map.
        Return the result in a dictionary where the key is the value and
        it maps to the count.
    """
    xsize = len(map)
    ysize = len(map[0])
    counts = { }
    for y in range(ysize):
        for x in range(xsize):
            if not allocated[x][y]:
                v = map[x][y]
                if not counts.has_key(v):
                    counts[v] = 0
                counts[v] = counts[v] + 1
    return counts    
   
def setChosen(source, allocated, sink, chosen, value):
    """
       For every position in source which has a value in chosen,
       set the corresponding value in sink to value. Also, set the
       corresponding position in allocated to 1.
    """
    xsize = len(source)                
    ysize = len(source[0])
    for y in range(ysize):
        for x in range(xsize):
            if source[x][y] in chosen:
                allocated[x][y] = 1
                sink[x][y] = value

def setTop(source, allocated, sink, value, percentage):
    """
        We need to set at least percentage of the locations in sink
        to value, but we can not use ones that are allocated.
        Choose the highest valued locations in source.
    """
    required = size(source) * percentage / 100
    heightCounts = countValues(source, allocated)
    heights = heightCounts.keys()
    heights.sort()
    heightsToUse = []
    tot = 0
    while tot < required:
        h = heights[-1]
        heights = heights[:-1]
        heightsToUse.append(h)
        c = heightCounts[h]
        tot = tot + c
    setChosen(source, allocated, sink, heightsToUse, value)
       

def setBottom(source, allocated, sink, value, percentage):
    """
        We need to set at least percentage of the locations in sink
        to value, but we can not use ones that are allocated.
        Choose the lowest valued locations in source.
    """
    required = size(source) * percentage / 100
    heightCounts = countValues(source, allocated)
    heights = heightCounts.keys()
    heights.sort()
    heightsToUse = []
    tot = 0
    while tot < required:
        h = heights[0]
        heights = heights[1:]
        heightsToUse.append(h)
        c = heightCounts[h]
        tot = tot + c
    setChosen(source, allocated, sink, heightsToUse, value)

def constmap(n):
    "Return a map which has only the value n in every position."
    return genmap(lambda x,y, n=n : n) 
    
def cloneMap(map):
    "Return an exact copy of a map."
    copy = []
    for col in map:    
        copy.append(col[:])
    return copy
    
def transform(map, matrix):
    """
        Apply a transformation to a map.
        The transformation, given in matrix, defines the influence that
        the neighbours of a cell have on the new value of that cell.
    """
    result = cloneMap(map)
    xmatsize = len(matrix)
    ymatsize = len(matrix[0])
    mxmin = xmatsize / 2
    mymin = ymatsize / 2
    xsize = len(map)                
    ysize = len(map[0])
    myrange = range(ymatsize)
    mxrange = range(xmatsize)
    yrange = range(ysize)
    xrange = range(xsize)
    for y in yrange:
        for x in xrange: 
               tot = 0
               weights = 0
               for xd in mxrange:
                   xc = x + xd - mxmin
                   if xc in xrange:
                       for yd in myrange:
                           yc = y + yd - mymin
                           if yc in yrange:
                               weight = matrix[xd][yd]
                               tot = tot + weight * map[xc][yc]
                               weights = weights + weight
               result[x][y] = tot/weights          
    return result   
    
def genSlopedMap(min, max):
    "A map that is like min at the top left and like max at the bottom right."
    lo = 0
    hi = XSIZE + YSIZE
    return genmap(lambda x, y, lo=lo, hi=hi: ((x+y) * hi + (hi-x-y) * lo) / hi)
    
def genBackSlopedMap(min, max):
    "A map that is like min at the top right and like max at the bottom left."
    return genmap(lambda x, y, lo=min, hi=max: backslope(x, y, lo=lo, hi=hi))
        
def backslope(x, y, lo, hi):
    "Helper function for genBackSlopedMap."
    diff = XSIZE + YSIZE
    h = y - x + XSIZE
    v = (h * lo + (diff - h) * hi) / diff
    return v
    
def genRandomMap(min, max):
    "Generate a map with uniformly random values from min to max inclusive."
    return genmap(lambda x, y, min=min, max=max : whrandom.randint(min,max)) 
    
def addMaps(m1, m2):
    "Sum maps m1 and m2. like standard matrix addition."
    result = cloneMap(m1)
    xsize = len(m1)
    ysize = len(m1[0])
    for y in range(ysize):
        for x in range(xsize):
            result[x][y] = m1[x][y] + m2[x][y] 
    return result    
    
def mapWhere(map, values):
    """
        Return a boolean map where a position is true if its value in map is in
        the list values.
    """
    result = cloneMap(map)    
    xsize = len(map)
    ysize = len(map[0])
    for y in range(ysize):
        for x in range(xsize):
            result[x][y] = map[x][y] in values 
    return result 
    
def inverse(map):    
    "Return the logical negation of a boolean map."  
    result = cloneMap(map)    
    xsize = len(map)
    ysize = len(map[0])
    for y in range(ysize):
        for x in range(xsize):
            result[x][y] = 1-map[x][y]
    return result 
    
def getCityTypes(map, x, y):
    return [map[x][y], map[x+1][y], map[x][y+1], map[x+1][y+1]]    
    
"""
Cities are NOT allowed to be placed on the bottom or right edges.
(e.g (99,39) is not allowed)
	
This is because cities need 4 maptiles and could
therefore be placed on maptiles that don't exist.
"""

def markAllocated(allocated, x, y, radius, diameter):
    xrange = range(len(allocated))
    yrange = range(len(allocated[0]))
    for xd in range(0, diameter):
        for yd in range(0, diameter):
            xc = x + xd - radius;
            yc = y + yd - radius;
            if xc in xrange and yc in yrange:  
                allocated[xc][yc] = 1  

def allocateCities(cityMap, map, allocated):
    "Place cities on the character map."
    required = size(map) / 100 * CITY_PERCENT
    xsize = len(map)
    ysize = len(map[0])   
    cities = 0
    while cities < required:
        x = whrandom.randint(0, xsize-2)
        y = whrandom.randint(0, ysize-2)
        # check not too close to another city
        isAllocated = getCityTypes(allocated, x, y)
        if 1 in isAllocated:
            continue
        # check we are not in an impossible place
        terrain = getCityTypes(map, x, y)
        #if MOUNTAIN in terrain or WATER in terrain or SWAMP in terrain:
        if WATER in terrain:
            continue
        # check no other building is here
        buildings = getCityTypes(cityMap, x, y)
        if TEMPLE in buildings or RUIN in buildings:
            continue     
        cityMap[x][y] = CITY
        cityMap[x+1][y] = PART_OF_CITY
        cityMap[x][y+1] = PART_OF_CITY
        cityMap[x+1][y+1] = PART_OF_CITY
        cities = cities + 1
        markAllocated(allocated, x, y, CITY_FAR_APART, CITY_FAR_APART * 2 + 2)  
        
def allocateRuins(cityMap, map, allocated):
    "Place ruins on the character map."
    required = size(map) / 100 * RUIN_PERCENT
    xsize = len(map)
    ysize = len(map[0])   
    ruins = 0
    while ruins < required:
        x = whrandom.randint(0, xsize-1)
        y = whrandom.randint(0, ysize-1)
        if allocated[x][y]:
            continue
        if map[x][y] in [WATER]:
            continue
        if cityMap[x][y] != NOTHING:
            continue
        cityMap[x][y] = RUIN
        ruins = ruins + 1
        markAllocated(allocated, x, y, RUIN_FAR_APART, RUIN_FAR_APART * 2 + 1) 
		    
def allocateTemples(cityMap, map, allocated):
    "Place temples on the character map."
    required = size(map) / 100 * TEMPLE_PERCENT
    xsize = len(map)
    ysize = len(map[0])     
    temples = 0
    while temples < required:
        x = whrandom.randint(0, xsize-1)
        y = whrandom.randint(0, ysize-1)
        if allocated[x][y]:
            continue
        if map[x][y] in [WATER]:
            continue
        if cityMap[x][y] != NOTHING:
            continue
        cityMap[x][y] = TEMPLE
        temples = temples + 1
        markAllocated(allocated, x, y, TEMPLE_FAR_APART, TEMPLE_FAR_APART * 2 + 1)    
       
def flipYAxis(map):
    "Reflect a map in its Y axis"
    result = cloneMap(map)    
    xsize = len(map)
    ysize = len(map[0])
    for y in range(ysize):
        for x in range(xsize):
            result[x][y] = map[xsize-1-x][y]
    return result             
    
def flipXAxis(map):
    "Reflect a map in its X axis"
    result = cloneMap(map)    
    xsize = len(map)
    ysize = len(map[0])
    for y in range(ysize):
        for x in range(xsize):
            result[x][y] = map[x][ysize-1-y]
    return result             
        
def transpose(map):
    "Reflect map along its NW/SE axis."
    result = cloneMap(map)    
    xsize = len(map)
    ysize = len(map[0])
    for y in range(ysize):
        for x in range(xsize):
            result[x][y] = map[y][x]
    return result                
    
def changeSymmetry(map1, map2):
    "Reflect/flip a map so it doesn't always have features in the same area."
    if whrandom.randint(0,1) == 1:
        map1 = flipXAxis(map1)  
        map2 = flipXAxis(map2)  
    if whrandom.randint(0,1) == 1:
        map1 = flipYAxis(map1) 
        map2 = flipYAxis(map2)
    if whrandom.randint(0,1) == 1:
        map1 = transpose(map1) 
        map2 = transpose(map2)          
    return map1, map2  
            
AVERAGE5 = [[1,1,1,1,1], [1,2,2,2,1], [1,2,3,2,1], [1,2,2,2,1], [1,1,1,1,1]]
AVERAGE3 = [[1,1,1], [1,2,1], [1,1,1]]

def generateMap():
    heightMap = addMaps(genSlopedMap(0,9), genRandomMap(0, 99)) 
    heightMap = transform(heightMap, AVERAGE5)
    
    pickle.dump(heightMap, open("stage.1", "w"))

def setTerrain():
    heightMap = pickle.load(open("stage.1", "r"))

    allocated = constmap(0)
    sink = constmap('.')
    setTop(heightMap, allocated, sink, MOUNTAIN, MOUNTAIN_PERCENT)
    setTop(heightMap, allocated, sink, HILL, HILL_PERCENT)
    setBottom(heightMap, allocated, sink, WATER, WATER_PERCENT)

    rainfallMap = addMaps(genBackSlopedMap(0,25), genRandomMap(0,69))
    rainfallMap = transform(rainfallMap, AVERAGE5)
    allocated = inverse(mapWhere(sink, PLAIN))
    setTop(rainfallMap, allocated, sink, FOREST, FOREST_PERCENT)
    setBottom(rainfallMap, allocated, sink, SWAMP, SWAMP_PERCENT)

    pickle.dump(sink, open("stage.2_0", "w"))
    pickle.dump(allocated, open("stage.2_1", "w"))

def setBuildings():
    sink = pickle.load(open("stage.2_0", "r"))
    allocated = pickle.load(open("stage.2_1", "r"))

    citySink = constmap(' ')
    allocated = constmap(0)
    allocateCities(citySink, sink, allocated)
    allocateRuins(citySink, sink, allocated)
    allocateTemples(citySink, sink, allocated)

    (sink, citySink) = changeSymmetry(sink, citySink)

    # write into file
    file = open('random.map', 'w')
    writeToFile(file, toStringMap(sink))
    writeToFile(file, toStringMap(citySink))

if len(sys.argv) == 1:
    generateMap()
    setTerrain()
    setBuildings()
    sys.exit(0)

if sys.argv[1] == "-h":
    print """
    Usage: ./genmap.py <stage>
    
      <stage>: stage0       Generate empty random map
               stage1       Set terrain
               stage2       Set buildings

      If called without a parameter it generates a complete Map in one step.
    """
    sys.exit(0)

if sys.argv[1] == "stage0":
    generateMap()
elif sys.argv[1] == "stage1":
    setTerrain()
elif sys.argv[1] == "stage2":
    setBuildings()
